# from ensurepip import version
import excel2sbol.compiler as e2s
import os
import json
from datetime import datetime

def converter(file_path_in, file_path_out, sbol_version=3, homespace="http://examples.org/", file_format=None,  username=None, password=None, url = None):
    """Convert a given excel file to SBOL

    Args:
        file_path_in (string): path to excel file
        file_path_out (string): desired path to sbol file
        sbol_version (int): sbol version number, defaults to 3
    """
    if username is not None and password is not None and url is not None:
        # print(username, password, url)
        os.environ["SBOL_USERNAME"] = username
        os.environ["SBOL_PASSWORD"] = password
        os.environ["SBOL_URL"] = url
        
    col_read_df, to_convert, compiled_sheets, version_info, homespace2 = e2s.initialise(file_path_in)
    dict = e2s.initialise_welcome(file_path_in)
    for key, value in dict.items():
        if isinstance(value, datetime):
            dict[key] = value.isoformat()
    if dict is not None:
        os.environ["SBOL_DICTIONARY"] = json.dumps(dict)
    # print(dict)

    if len(homespace2) > 0:
        homespace = homespace2
        print(f'Conversion will happen with homespace {homespace} as specified in the excel sheet')

    sbol_version = version_info
    print(f'Conversion will happen with sbol version {sbol_version} as specified in the excel sheet')

    if sbol_version == 2:
        doc, dict_of_objs, sht_convert_dict = e2s.parse_objects(col_read_df,
                                                                to_convert,
                                                                compiled_sheets,
                                                                homespace)
    elif sbol_version == 3:
        doc, dict_of_objs, sht_convert_dict = e2s.parse_objects3(col_read_df,
                                                                 to_convert,
                                                                 compiled_sheets,
                                                                 homespace)

    e2s.column_parse(to_convert, compiled_sheets, sht_convert_dict,
                     dict_of_objs, col_read_df, doc, file_path_out,
                     sbol_version=sbol_version, file_format=file_format)
   
    


import pandas as pd
import excel_sbol_utils.helpers as hf
import excel2sbol.lookup_compiler as lk
import excel2sbol.comp_column_functions2 as cf2
import logging
import sbol2
import sbol3
import math
import re
import os

# the homespace only works if the change is made to pysbol2 shown in https://github.com/SynBioDex/pySBOL2/pull/411/files

def initialise_welcome(file_path_in):
    init_info = pd.read_excel(file_path_in, sheet_name="Init",
                              skiprows=9, index_col=0,
                              engine='openpyxl')
    init_info = init_info.applymap(lambda x: x.strip() if isinstance(x, str) else x).to_dict('index')

    version_info = pd.read_excel(file_path_in, sheet_name="Init",
                                 nrows=4, index_col=0, header=None,
                                 engine='openpyxl')
    version_info = version_info.applymap(lambda x: x.strip() if isinstance(x, str) else x).to_dict('index')
    if 'Homespace' in version_info:
        homespace = version_info['Homespace'][1]
    else:
        homespace = ""
    version_info = version_info['SBOL Version'][1]

    # For key in dict read in sheet,
    # if sheet convert = true, add to convert list
    compiled_sheets = {}
    to_convert = []
    for sheet_name, val in init_info.items():
        # print(f"reading in {sheet_name}...")
        # MY CODE
        if sheet_name.lower() == "welcome":
            print("Processing the welcome page...")
            try:
                welcome_metadata = pd.read_excel(file_path_in, sheet_name=sheet_name,
                                                 index_col=0, engine='openpyxl').fillna("")
                print("Welcome Page Metadata:")
                
                dict = {}
                for _, row in welcome_metadata.iterrows():
                    if len(row) >= 2:
                        key, value = row[0], row[1]
                        if isinstance(key, str) and key.strip():
                            dict[key.strip()] = value.strip() if isinstance(value, str) else value
                for key, value in dict.items():
                    print(f"{key}: {value}")
                return dict
                
            except Exception as e:
                print(f"Error reading the welcome page: {e}")
                return None
            
        else:
            return
    
def initialise(file_path_in):
    init_info = pd.read_excel(file_path_in, sheet_name="Init",
                              skiprows=9, index_col=0,
                              engine='openpyxl')
    init_info = init_info.applymap(lambda x: x.strip() if isinstance(x, str) else x).to_dict('index')

    version_info = pd.read_excel(file_path_in, sheet_name="Init",
                                 nrows=4, index_col=0, header=None,
                                 engine='openpyxl')
    version_info = version_info.applymap(lambda x: x.strip() if isinstance(x, str) else x).to_dict('index')
    if 'Homespace' in version_info:
        homespace = version_info['Homespace'][1]
    else:
        homespace = ""
    version_info = version_info['SBOL Version'][1]

    # For key in dict read in sheet,
    # if sheet convert = true, add to convert list
    compiled_sheets = {}
    to_convert = []
    for sheet_name, val in init_info.items():
        # print(f"reading in {sheet_name}...")
        # MY CODE
        if sheet_name.lower() == "welcome":
            continue
		# MY CODE
        
        convert = val['Convert']

        if convert:
            to_convert.append(sheet_name.strip())

        # read in collections, description, library
        sheet_dict = {}

        if val['Has Collections']:
            x = val['Collect Cols']
            x = x.split(',')
            x = [int(i) for i in x]

            sheet_dict['collection_info'] = pd.read_excel(file_path_in, sheet_name=sheet_name,
                                                          header=None,
                                                          nrows=val['# of Collect Rows'],
                                                          usecols=x,
                                                          index_col=0,
                                                          engine='openpyxl').to_dict('index')
        else:
            sheet_dict['collection_info'] = {}

        if val['Has Descripts']:
            x = val['Descript Cols']
            if isinstance(x, (float, int)):
                x = int(x)
                x = [x]
            elif isinstance(x, str):
                x = x.split(',')
                x = [int(i) for i in x]

            sheet_dict['description'] = pd.read_excel(file_path_in,
                                                      sheet_name=sheet_name,
                                                      header=None,
                                                      skiprows=int(val['Descript Start Row']),
                                                      nrows=1,
                                                      usecols=x,
                                                      engine='openpyxl').iloc[0, 0]
        else:
            sheet_dict['description'] = ""

        skipval = val['Lib Start Row'] - 1  #to avoid zero index confusion
        lib_df = pd.read_excel(file_path_in, sheet_name=sheet_name,
                               header=0, skiprows=skipval,
                               engine='openpyxl').fillna("")
        sheet_dict['library'] = lib_df.applymap(lambda x: x.strip() if isinstance(x, str) else x).to_dict('list')

        # need dicitonary with as keys every column name and as values a list of values (note ordered list and need place holder empty values)
        compiled_sheets[sheet_name] = sheet_dict

    # read in column_dict sheet
    col_read_df = pd.read_excel(file_path_in,
                                sheet_name="column_definitions", header=0,
                                engine='openpyxl')
    col_read_df = col_read_df.applymap(lambda x: x.strip() if isinstance(x, str) else x)

    # processing to turn init columns into 'sheet' columns
    # pull all column names
    extra_cols = list(list(init_info.values())[0].keys())
    extra_cols = extra_cols[8:]
    for conv_sht in to_convert:
        for xcol in extra_cols:
            init_val = init_info[conv_sht][xcol]
            if isinstance(init_val, str) or not math.isnan(init_val):
                # add row to col def sheet
                new_row = col_read_df[col_read_df['Sheet Name'] == 'Init']
                new_row = new_row[new_row['Column Name'] == xcol].to_dict('records')
                new_row[0]['Sheet Name'] = conv_sht
                new_row_df = pd.DataFrame(new_row)
                # col_read_df = col_read_df.append(new_row_df)
                col_read_df = pd.concat([col_read_df, new_row_df])

                # add col to compiled_sheets
                num_rows = len(list(compiled_sheets[conv_sht]['library'].values())[0])
                val_list = [init_val for x in range(num_rows)]  # make a list of appropriate length
                compiled_sheets[conv_sht]['library'][xcol] = val_list

    # re index as otherwise causes issues later
    col_read_df = col_read_df.reset_index(drop=True)
    return(col_read_df, to_convert, compiled_sheets, version_info, homespace)


def parse_objects(col_read_df, to_convert, compiled_sheets,
                  homespace='http://examples.org/', sbol_version=2):
    """Making a list of all objects in the document"""

    # create uris for every item in to convert sheets
    # (note might want generic top level
    # if object type is not an sbol object type)

    dict_of_objs = {}
    sht_convert_dict = {}
    doc = sbol2.Document()
    sbol2.setHomespace(homespace)

    # sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_COMPLIANT_URIS, False)
    sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, False)

    for sht in to_convert:
        sht_df = col_read_df.loc[col_read_df['Sheet Name'] == sht]

        try:
            dis_name_col = sht_df.loc[col_read_df['SBOL Term'] == 'sbol_displayId']['Column Name'].values[0]
        except IndexError as e:
            raise KeyError(f'The sheet "{sht}" has no column with sbol_displayId as type. Thus the following error was raised: {e}')

        try:
            obj_type_col = sht_df.loc[col_read_df['SBOL Term'] == 'sbol_objectType']['Column Name'].values[0]
        except IndexError as e:
            raise KeyError(f'The sheet "{sht}" has no column with sbol_objectType as type. Thus the following error was raised: {e}')

        sht_convert_dict[sht] = dis_name_col
        ids = compiled_sheets[sht]['library'][dis_name_col]
        types = compiled_sheets[sht]['library'][obj_type_col]

        for ind, id in enumerate(ids):
            sanitised_id = hf.check_name(id)
            uri = f'{sbol2.getHomespace()}{sanitised_id}'

            if hasattr(sbol2, types[ind]):
                varfunc = getattr(sbol2, types[ind])
                if types[ind] == "CombinatorialDerivation":
                    # print('combdev', sanitised_id, types[ind])
                    # template = sbol2.ComponentDefinition(f'{sanitised_id}_template')
                    # template.displayId = f'{sanitised_id}_template'
                    # dict_of_objs[f'{sanitised_id}_template'] = {'uri': f'{sbol2.getHomespace()}{sanitised_id}_template',
                    #                                             'object': template, 'displayId': f'{sanitised_id}_template'}
                    obj = varfunc(uri=sanitised_id)
                    # print(f'here, {sanitised_id}')
                else:
                    # print(sanitised_id, types[ind])
                    obj = varfunc(sanitised_id)
                obj.displayId = sanitised_id
                # if "Supplement" in obj.displayId:
                #     print(obj, type(obj))

            else:
                # if not a known sbol class use generic toplevel
                obj = sbol2.TopLevel(type_uri=types[ind], uri=uri, version='1')

            dict_of_objs[id] = {'uri': uri, 'object': obj,
                                'displayId': sanitised_id}

    for obj_name in dict_of_objs:
        obj = dict_of_objs[obj_name]['object']
        # print(obj_name, obj, type(obj))
        doc.add(obj)
    return(doc, dict_of_objs, sht_convert_dict)


def parse_objects3(col_read_df, to_convert, compiled_sheets,
                   homespace='http://examples.org', sbol_version=3):
    """Making a list of all objects in the document"""

    # create uris for every item in to convert sheets
    # (note might want generic top level
    # if object type is not an sbol object type)

    dict_of_objs = {}
    sht_convert_dict = {}
    doc = sbol3.Document()
    sbol3.set_namespace(homespace)

    # sbol3.ConfigOptions.SBOL_TYPED_URIS = False
    # sbol3.Config.setOption(sbol3.ConfigOptions.SBOL_TYPED_URIS = False)
    # sbol3.Config.setOption(sbol3.ConfigOptions.SBOL_TYPED_URIS, False)

    for sht in to_convert:
        sht_df = col_read_df.loc[col_read_df['Sheet Name'] == sht]

        try:
            dis_name_col = sht_df.loc[col_read_df['SBOL Term'] == 'sbol_displayId']['Column Name'].values[0]
        except IndexError as e:
            raise KeyError(f'The sheet "{sht}" has no column with sbol_displayID as type. Thus the following error was raised: {e}')

        try:
            obj_type_col = sht_df.loc[col_read_df['SBOL Term'] == 'sbol_objectType']['Column Name'].values[0]
        except IndexError as e:
            raise KeyError(f'The sheet "{sht}" has no column with sbol_objectType as type. Thus the following error was raised: {e}')

        try:
            mol_type_col = sht_df.loc[col_read_df['SBOL Term'] == 'sbol_types']['Column Name'].values[0]
            mol_types = compiled_sheets[sht]['library'][mol_type_col]
        except IndexError:
            mol_types = None

        sht_convert_dict[sht] = dis_name_col
        ids = compiled_sheets[sht]['library'][dis_name_col]
        obj_types = compiled_sheets[sht]['library'][obj_type_col]

        for ind, id in enumerate(ids):
            sanitised_id = hf.check_name(id)

            uri = f'{sbol3.get_namespace()}/{sanitised_id}'

            if hasattr(sbol3, obj_types[ind]):
                varfunc = getattr(sbol3, obj_types[ind])
                if obj_types[ind] == "Component":
                    # checks that a molecule type is given and it isn't a boolean like circular
                    if mol_types is not None and isinstance(mol_types[ind], str):
                        obj = varfunc(f'{homespace}/{sanitised_id}', mol_types[ind])
                    else:
                        obj = varfunc(sanitised_id, sbol3.SBO_DNA)
                        logging.warning(f'As no molecule type was giving the component {id} was initiated as a DNA molecule')
                elif obj_types[ind] == "CombinatorialDerivation":
                    template = sbol3.Component(f'{homespace}/{sanitised_id}_template', sbol3.SBO_DNA)
                    template.displayId = f'{sanitised_id}_template'
                    dict_of_objs[f'{sanitised_id}_template'] = {'uri': f'{homespace}/{sanitised_id}_template',
                                                                'object': template, 'displayId': f'{sanitised_id}_template'}

                    obj = varfunc(sanitised_id, template)
                    # doesnt work for comb dev at the moment!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                else:
                    obj = varfunc(f'{homespace}/{sanitised_id}')
                obj.displayId = sanitised_id

            else:
                # if not a known sbol class use generic toplevel
                obj = sbol3.TopLevel(type_uri=obj_types[ind], identity=uri)

            dict_of_objs[id] = {'uri': uri, 'object': obj,
                                'displayId': sanitised_id}

    for obj_name in dict_of_objs:
        obj = dict_of_objs[obj_name]['object']
        doc.add(obj)
    return(doc, dict_of_objs, sht_convert_dict)


class TermClass:
    def __init__(self, row):
        self.row_num = row


def column_parse(to_convert, compiled_sheets, sht_convert_dict, dict_of_objs,
                 col_read_df, doc, file_path_out, sbol_version=3, file_format=None):
    doc_pref_terms = ['rdf', 'rdfs', 'xsd', 'sbol']
    data_source_id_to_update = {}

    for sht in to_convert:
        print(sht)
        sht_lib = compiled_sheets[sht]['library']

        # pulls first column and checks the number of elements in it
        num_rows = len(sht_lib[list(sht_lib.keys())[0]])

        for row_num in range(0, num_rows):

            term_dict = TermClass(row_num)

            disp_id = sht_lib[sht_convert_dict[sht]][row_num]
            obj = dict_of_objs[disp_id]['object']
            obj_uri = dict_of_objs[disp_id]['uri']

            for col in sht_lib.keys():
                cell_val = sht_lib[col][row_num]

                if cell_val != '':
                    # checks that the cell isn't blank
                    col_convert_df = col_read_df.loc[(col_read_df['Sheet Name'] == sht) & (col_read_df['Column Name'] == col)]
                    if col_convert_df.empty:
                        raise ValueError(f"There is an issue with the column definitions sheet missing values. Sheet:'{sht}' with Column:'{col}' cannot be found. Please check for any spaces.")

                    # split method
                    split_on = col_convert_df['Split On'].values[0]
                    split_on = split_on.split('"')
                    split_on = [x for x in split_on if x != '']
                    split_on = '[' + "".join(split_on) + ']'

                    # used as string will always be '[]' at least
                    if len(split_on) > 2:
                        cell_val = re.split(split_on, cell_val)
                    if isinstance(cell_val, list):
                        cell_val = [x.strip() for x in cell_val]

                    # cell value or list of cell values based on lookups
                    if isinstance(cell_val, list):
                        for ind, val in enumerate(cell_val):
                            cell_val[ind] = lk.up(col_convert_df, val,
                                                  compiled_sheets,
                                                  dict_of_objs)
                    else:
                        cell_val = lk.up(col_convert_df, cell_val,
                                         compiled_sheets, dict_of_objs)

                    # if converted to empty cell or
                    # empty string then skip the rest
                    is_nan = False
                    if isinstance(cell_val, float):
                        is_nan = math.isnan(cell_val)
                    if cell_val == "" or is_nan:
                        continue

                    # Ensures that the cell value after possible conversion
                    # matches one of the patterns specified
                    pattern = col_convert_df['Pattern'].values[0]
                    if isinstance(pattern, str) and len(pattern) > 2:
                        pattern = pattern = pattern.split('"')
                        pattern = [x for x in pattern if x != '' and x != ' ']
                        if isinstance(cell_val, list):
                            for val in cell_val:
                                pat_truth = [re.match(pat, val) for pat in pattern]
                                pat_truth = [True for pat in pat_truth if pat is not None]
                                if len(pat_truth) < 1:
                                    raise ValueError(f'The cell value provided did not meet (any of) the pattern criteria, cell value: {val} (in sheet:{sht}, column:{col},  row:{disp_id}), pattern:{pattern}')
                        else:
                            pat_truth = [re.match(pat, cell_val) for pat in pattern]
                            pat_truth = [True for pat in pat_truth if pat is not None]
                            if len(pat_truth) < 1:
                                raise ValueError(f'The cell value provided did not meet (any of) the pattern criteria, cell value: {cell_val} (in sheet:{sht}, column:{col},  row:{disp_id}), pattern:{pattern}')

                    # carry out method of column processing based on
                    # the sbol_term of the column
                    # This creates an object with the converted cell values
                    # hierarchy: sbol term, multicolumn, column name, cell val
                    mcol = col_convert_df['Multicolumn'].values[0]
                    sbol_term = col_convert_df['SBOL Term'].values[0]

                    if hasattr(term_dict, sbol_term):
                        sbol_dict = getattr(term_dict, sbol_term)
                    else:
                        sbol_dict = {}

                    if isinstance(mcol, str):
                        if mcol not in sbol_dict:
                            sbol_dict[mcol] = {}
                        sbol_dict[mcol][col] = cell_val
                    else:
                        sbol_dict[col] = cell_val

                    setattr(term_dict, sbol_term, sbol_dict)


            # print(term_dict.__dict__)
            for term in term_dict.__dict__:
                if term != 'row_num':
                    # print(term, getattr(term_dict, term))
                    col_cell_dict = getattr(term_dict, term)
                    term_coldef_df = col_read_df[(col_read_df['SBOL Term'] == term) & (col_read_df['Sheet Name'] == sht)]
                    rj = cf2.rowobj(obj, obj_uri, dict_of_objs, doc,
                                    col_cell_dict, sht, disp_id,
                                    term_coldef_df, doc_pref_terms, data_source_id_to_update)
                    sw = cf2.switch1()
                    sw.switch(rj, term, sbol_version)
                    doc_pref_terms = rj.doc_pref_terms
                    data_source_id_to_update = rj.data_source_id_to_update
            # print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")

    # this makes sure all references to an object are also updated (as well as any derived terms)
    if len(data_source_id_to_update) > 0:
        doc = hf.update_uri_refs(doc, data_source_id_to_update, use_derived=True, derived_ls = ['_sequence'])

    
    if os.getenv("COUNTER") == "Error found":
        return
    elif file_format is None:
        doc.write(file_path_out)
    else:
        doc.write(file_path_out, file_format = file_format)
    
    return


from unittest import TestCase
import pytest
import excel2sbol.converter as confun
import excel2sbol.compiler as e2s
import os
import tempfile
import rdflib
import rdflib.compare
import openpyxl
import sbol3

TESTFILE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test_files')

def test_conversion():
    sbol3.set_namespace('http://examples.org')

    # Convert the document into a temporary file
    # From that file have a document read the file
    # Use that document to test

    file_path_in = os.path.join(TESTFILE_DIR, 'simple_library2.xlsx')

    with tempfile.TemporaryDirectory() as dirpath:
        file_path_out = os.path.join(dirpath, 'sample_out.xml')
        confun.converter(file_path_in, file_path_out, 3)

        doc = sbol3.Document()
        doc.read(file_path_out)

        # Tests ensuring there are x amount of objects in different sheet collections
        assert len(doc.find('Composite_u32_Parts').members) == 6
        assert len(doc.find('Basic_u32_Parts').members) == 26

        # tests for later implementation:
        # assert len(doc.find('LinearDNAProducts').members) == 2
        # assert len(doc.find('FinalProducts').members) == 2

        # Holistic test here
        # These RDF files are made isomorphic in order to compare them 1:1
        # When isomorphic, they are aligned in the same way to easily compare.

        expected = os.path.join(TESTFILE_DIR,
                                    'sample_out.xml')
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected)
        expected_iso = rdflib.compare.to_isomorphic(expected_graph)
        output_graph = rdflib.Graph()
        output_graph.parse(file_path_out)
        output_iso = rdflib.compare.to_isomorphic(output_graph)

        # Check to see if the graphs (SBOL files) are equivalent
        assert output_iso.__eq__(expected_iso)

# def test_constraints(self):
#    sbol3.set_namespace('http://examples.org')




# Make sure to update the dependency graphic
# https://github.com/SynBioDex/Excel-to-SBOL/blob/master/images/dependency_structure.PNG
# if change are made to modle dependencies
import sbol3
import sbol2
from inspect import getmembers, isfunction
import excel_sbol_utils.library3 as exutil3
import excel_sbol_utils.library2 as exutil2


class rowobj():

    def __init__(self, obj, obj_uri, obj_dict, doc, col_cell_dict,
                 sheet, display_id, term_coldef_df,  doc_pref_terms, data_source_id_to_update):
        self.obj = obj
        self.obj_uri = obj_uri
        self.obj_dict = obj_dict
        self.doc = doc
        self.sheet = sheet
        self.col_cell_dict = col_cell_dict
        self.sht_row = display_id
        self.term_coldef_df = term_coldef_df
        self.doc_pref_terms = doc_pref_terms
        self.data_source_id_to_update = data_source_id_to_update


class switch1():
    func_list2 = func_list = [o[0] for o in getmembers(exutil2) if isfunction(o[1])]
    func_list3 = func_list = [o[0] for o in getmembers(exutil3) if isfunction(o[1])]

    def switch(self, rowobj, sbol_term, sbol_version):
        if sbol_version == 2:
            self.func_list = self.func_list2
            exutil = exutil2
        elif sbol_version == 3:
            self.func_list = self.func_list3
            exutil = exutil3
        else:
            raise ValueError(f"SBOL Version ({sbol_version}) given to switch has not been implemented yet")

        # split sbol term into prefix and suffix
        self.sbol_term = sbol_term
        self.sbol_term_pref = sbol_term.split("_", 1)[0]
        try:
            self.sbol_term_suf = sbol_term.split("_", 1)[1]
        except IndexError:
            raise ValueError(f"The SBOL Term '{sbol_term}' (sheet name: {self.sheet}) does not appear to have an underscore")

        # if not applicable then do nothing
        if sbol_term == "Not_applicable":
            pass

        # if a special function has been defined in excel-sbol-utils then use that
        elif self.sbol_term_suf in self.func_list:
            return getattr(exutil, self.sbol_term_suf)(rowobj)

        # if it is an sbol term use standard pySBOL implementation
        # unless it is a top level object in which case the standard
        # implementations don't work
        elif self.sbol_term_pref == "sbol":

            for col in rowobj.col_cell_dict:

                cell_val = rowobj.col_cell_dict[col]

                parental_lookup = rowobj.term_coldef_df[rowobj.term_coldef_df['Column Name'] == col]['Parent Lookup']
                if parental_lookup.values[0]:
                    # switches the object being worked on
                    rowobj.obj = rowobj.obj_dict[cell_val]['object']
                    cell_val = rowobj.obj_uri

                if hasattr(rowobj.obj, self.sbol_term_suf):
                    # if the attribute is a list append the new value
                    if isinstance(getattr(rowobj.obj, self.sbol_term_suf), list):
                        current = getattr(rowobj.obj, self.sbol_term_suf)
                        # if the col_cell_dict has multiple columns append each

                        # if cell_val is a dict then mcol must have been given
                        if isinstance(cell_val, dict):
                            raise TypeError(f"A multicolumn value was unexpectedly given for sheet:{rowobj.sheet}, row:{rowobj.sht_row}, sbol term :{self.sbol_term}, sbol term dict: {rowobj.col_cell_dict}")
                        # if the cell_val is a list append the whole list
                        elif isinstance(cell_val, list):
                            setattr(rowobj.obj, self.sbol_term_suf, current + cell_val)
                        # otherwise append as a list object
                        else:
                            setattr(rowobj.obj, self.sbol_term_suf, current + [cell_val])

                    # if type sbol list then add by special append
                    # rather than regular list append
                    elif isinstance(getattr(rowobj.obj, self.sbol_term_suf), sbol3.refobj_property.ReferencedObjectList):

                        # if cell_val is a dict then mcol must have been given
                        if isinstance(cell_val, dict):
                            raise TypeError(f"A multicolumn value was unexpectedly given for sheet:{rowobj.sheet}, row:{rowobj.sht_row}, sbol term :{self.sbol_term}, sbol term dict: {rowobj.col_cell_dict}")
                        # else should be fine to append
                        else:
                            getattr(getattr(rowobj.obj, self.sbol_term_suf), 'append')(cell_val)

                    else:

                        # no iteration over list as else suggests that the property
                        # can't have multiple values

                        setattr(rowobj.obj, self.sbol_term_suf, cell_val)
                else:
                    raise ValueError(f'This SBOL object ({type(rowobj.obj)}) has no attribute {self.sbol_term_suf} (sheet:{rowobj.sheet}, row:{rowobj.sht_row}, sbol term dict:{rowobj.col_cell_dict})')

        else:
            # logging.warning(f'This sbol term ({self.sbol_term}) has not yet been implemented so it has been added via the default method')
            # define a new namespace if needed
            for col in rowobj.col_cell_dict:
                cell_val = rowobj.col_cell_dict[col]
                col_coldef_df = rowobj.term_coldef_df[rowobj.term_coldef_df['Column Name'] == col]

                if len(col_coldef_df) == 0:
                    raise TypeError(f"A multicolumn value was unexpectedly given for sheet:{rowobj.sheet}, row:{rowobj.sht_row}, sbol term :{self.sbol_term}, sbol term dict: {rowobj.col_cell_dict}")

                parental_lookup = col_coldef_df['Parent Lookup'].values[0]
                if parental_lookup:
                    # switches the object being worked on
                    rowobj.obj = rowobj.obj_dict[cell_val]['object']
                    cell_val = rowobj.obj_uri

                namespace_url = col_coldef_df['Namespace URL'].values[0]
                if self.sbol_term_pref not in rowobj.doc_pref_terms:
                    rowobj.doc.addNamespace(namespace_url, self.sbol_term_pref)
                    rowobj.doc_pref_terms.append(self.sbol_term_pref)

                col_type = col_coldef_df['Type'].values[0]
                # if type is uri make it a uri property
                if col_type == "URI":
                    # * allows multiple instance of this property
                    if not hasattr(rowobj.obj, self.sbol_term_suf):
                        if sbol_version == 2:
                            setattr(rowobj.obj, self.sbol_term_suf,
                                sbol2.URIProperty(rowobj.obj,
                                                f'{namespace_url}{self.sbol_term_suf}',
                                                '0', '*', []))
                            setattr(rowobj.obj, self.sbol_term_suf, cell_val)
                        elif sbol_version == 3:
                            setattr(rowobj.obj, self.sbol_term_suf,
                                sbol3.URIProperty(rowobj.obj,
                                                  f'{namespace_url}{self.sbol_term_suf}',
                                                  '0', '*', initial_value=[cell_val]))
                    else:
                        if not isinstance(cell_val, list):
                            cell_val = [cell_val]
                        current = getattr(rowobj.obj, self.sbol_term_suf)
                        setattr(rowobj.obj, self.sbol_term_suf, list(current) + cell_val)

                # otherwise implement as text property
                else:
                    # * allows multiple instance of this property
                    if not hasattr(rowobj.obj, self.sbol_term_suf):
                        if sbol_version == 2:
                            setattr(rowobj.obj, self.sbol_term_suf,
                                sbol2.TextProperty(rowobj.obj,
                                                f'{namespace_url}{self.sbol_term_suf}',
                                                '0', '*'))
                            setattr(rowobj.obj, self.sbol_term_suf, str(cell_val))
                        elif sbol_version == 3:
                            setattr(rowobj.obj, self.sbol_term_suf,
                                sbol3.TextProperty(rowobj.obj,
                                                   f'{namespace_url}{self.sbol_term_suf}',
                                                   '0', '*', initial_value=str(cell_val)))
                    else:
                        if not isinstance(cell_val, list):
                            cell_val = [cell_val]
                        current = getattr(rowobj.obj, self.sbol_term_suf)
                        setattr(rowobj.obj, self.sbol_term_suf, list(current) + cell_val)



# # import sbol3
import sbol2
import excel2sbol.helper_functions as hf
import os
import re


name = "hello world!1a"
print(hf.check_name(name))

# cwd = os.getcwd()
# file_path_out = os.path.join(cwd, "excel2sbol", 'tests', 'test_files', 'ACS_sbol_parts.xml')

# comps = ["A, B, C", "D", "E"]

# doc = sbol2.Document()

# obja = sbol2.ComponentDefinition("A")
# objb = sbol2.ComponentDefinition("B")
# objc = sbol2.ComponentDefinition("C")
# objd = sbol2.ComponentDefinition("D")
# obje = sbol2.ComponentDefinition("E")

# doc.add(obja)
# doc.add(objb)
# doc.add(objc)
# doc.add(objd)
# doc.add(obje)

# comdev = sbol2.CombinatorialDerivation("comdev")
# doc.add(comdev)

# comp_ind = 0
# variant_comps = {}
# for ind, comp in enumerate(comps):
#     if "," in comp or type(comdev) == sbol2.combinatorialderivation.CombinatorialDerivation:
#         comps[ind] = f'{comdev.displayId}_subcomponent_{comp_ind}'
#         uri = f'{comdev.displayId}_subcomponent_{comp_ind}'
#         sub_comp = sbol2.ComponentDefinition(uri)
#         sub_comp.displayId = f'{comdev.displayId}_subcomponent_{comp_ind}'
#         doc.add(sub_comp)
#         variant_comps[f'subcomponent_{comp_ind}'] = {'object': sub_comp, 'variant_list': comp}
#         comp_ind += 1
#     else:
#         comps[ind] = hf.check_name(comps[ind])

# temp = sbol2.ComponentDefinition("comdev_template")
# temp.displayId = f'{comdev.displayId}_template'
# doc.add(temp)

# temp.assemblePrimaryStructure(comps)
# # temp.compile(assembly_method=None)

# comdev.masterTemplate = temp
# for var in variant_comps:
#     #var = hf.check_name(var)
#     var_comp = sbol2.VariableComponent(f'var_{var}')
#     var_comp.displayId = f'var_{var}'
#     var_comp.variable = variant_comps[var]['object']

#     var_list = re.split(",", variant_comps[var]['variant_list'])
#     var_list = [f'{sbol2.getHomespace()}/{hf.check_name(x.strip())}' for x in var_list]
#     var_comp.variants = var_list
#     comdev.variableComponents.add(var_comp)


# doc.write(file_path_out)





# # doc = sbol3.Document()

# # colec = sbol3.Collection('FinalProducts', name='FinalProducts')
# # doc.add(colec)

# # # print(sbol3.get_namespace())

# # sbol_objs = doc.objects
# # sbol_objs_names = [x.name for x in sbol_objs]
# # if 'FinalProducts' not in sbol_objs_names:
# #     colec = sbol3.Collection('FinalProducts', name='FinalProducts')
# #     # colec.members.append('test')
# #     doc.add(colec)
# # else:
# #     colec = sbol_objs[sbol_objs_names.index('FinalProducts')]

# # colec.members.append('this')

# sbol2.CombinatorialDerivation()
# sbol2.ComponentDefinition()
# sbol2.getHomespace()
# sbol2.URIRef




# run by typing 'pytest' into the terminal
# for more detailed output use 'pytest -v -s'
import pytest
import excel2sbol.column_functions as cf
import os


class Test_sbol_methods:

    @pytest.mark.parametrize(
        'sbol_term, nm_url, cell_val, raising_err, expected',
        [
            ("Not_applicable", 'nm_url', 'cell_val', False, 'fake_NA'),
            ("add_new", 'nm_url', 'cell_val', False, 'fake_add_new')
        ]
    )
    def test_switch(self, sbol_term, nm_url, cell_val, raising_err, expected,
                    sbol_component, sbol_doc, monkeypatch):

        def fake_NA(self):
            return "fake_NA"

        def fake_add_new(self):
            return "fake_add_new"

        monkeypatch.setattr(cf.sbol_methods, 'Not_applicable', fake_NA)
        monkeypatch.setattr(cf.sbol_methods, 'add_new', fake_add_new)

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)
        if raising_err:
            with pytest.raises(expected):
                x.switch(sbol_term)
        else:
            assert x.switch(sbol_term) == expected

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, expected',
        [
            ('nm_url', 'atgcgcgcgc', False, 'atgcgcgcgc'),
            ('nm_url', 'atGC GCA', False, 'atgcgca'),
            ('nm_url', 'at1gcgcgcgc', True, TypeError),
            ('nm_url', 5, True, TypeError),
            # ('nm_url', 'attgz', True, TypeError),
            ('nm_url', 'ättgc', True, TypeError),
            ('nm_url', 'atuuc', False, 'atuuc')
        ]
    )
    def test_sbol_sequence(self, nm_url, cell_val, raising_err,
                           expected, sbol_component, sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)

        if raising_err:
            with pytest.raises(expected):
                x.sbol_sequence()

        else:
            x.sbol_sequence()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions[f'http://examples.org/{sbol_component.name}/1']
            seq = sbol_doc.sequences[f'http://examples.org/{sbol_component.name}_sequence/1']

            assert seq.elements == expected and str(comp.sequence) == f'http://examples.org/{sbol_component.name}_sequence/1'

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, expected',
        [
            ('nm_url', 'Any 1 descript with wεird stuff!', False,
             'Any 1 descript with wεird stuff!'),
            ('nm_url', 'Any odd description', False, 'Any odd description'),
            ('nm_url', 5, True, TypeError),
            ('nm_url', '5005', True, TypeError)
        ]
    )
    def test_dcterms_description(self, nm_url, cell_val,
                                 raising_err, expected, sbol_component,
                                 sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)
        if raising_err:
            with pytest.raises(expected):
                x.dcterms_description()

        else:
            x.dcterms_description()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions[f'http://examples.org/{sbol_component.name}/1']
            assert comp.description == expected

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, expected1, expected2',
        [
            ('nm_url', "testname", False, "testname", "testname"),
            ('nm_url', "test_name", False, "test_name", "test_name"),
            ('nm_url', "_test_name", False, "_test_name", "_test_name"),
            ('nm_url', "test_name12", False, "test_name12", "test_name12"),
            ('nm_url', "12test_name", False, "12test_name", "_12test_name"),
            ('nm_url', "test_name%", False, "test_name%", "test_name_u37_"),
            ('nm_url', "test_näme", False, "test_näme", "test_n_u228_me"),
            ('nm_url', "tεst_name", False, "tεst_name", "t_u949_st_name"),
            ('nm_url', 576, True, TypeError, "")
        ]
    )
    def test_sbol_displayId(self, nm_url, cell_val,
                            raising_err, expected1, expected2,
                            sbol_component, sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)
        if raising_err:
            with pytest.raises(expected1):
                x.sbol_displayId()

        else:
            x.sbol_displayId()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions['http://examples.org/test_comp_name/1']
            assert comp.name == expected1 and comp.displayId == expected2

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, roles, expected',
        [
            ('nm_url', False, False, ["0000316", '0000167'],
             ['0000316', '0000167']),
            ('nm_url', False, False, [], []),
            ('nm_url', True, False, ["0000316"], ['0000988', '0000316']),
            ('nm_url', True, False, [], ['0000988']),
            ('nm_url', "False", False, ["0000316"], ['0000316']),
            ('nm_url', "FALSE", False, [], []),
            ('nm_url', "TrUe", False, ["0000316"], ['0000988', '0000316']),
            ('nm_url', "TRUE", False, [], ['0000988']),
            ('nm_url', 0, False, ["0000316"], ['0000316']),
            ('nm_url', 1, False, [], ['0000988']),
            ('nm_url', "random stuff", True, ["0000316"], TypeError),
            ('nm_url', 576, True, [], TypeError)
        ]
    )
    def test_sbol_roleCircular(self, nm_url, cell_val, raising_err, roles,
                               expected, sbol_component, sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)
        if len(roles) > 0:
            for role in roles:
                sbol_component.roles = sbol_component.roles + [f'http://identifiers.org/so/SO:{role}']

        if raising_err:
            with pytest.raises(expected):
                x.sbol_roleCircular()

        else:
            expected = [f'http://identifiers.org/so/SO:{role}' for role in expected]
            x.sbol_roleCircular()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions['http://examples.org/test_comp_name/1']
            assert set(comp.roles) == set(expected)

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, roles, expected',
        [
            ('nm_url', "http://purl.obolibrary.org/obo/SO_0000167",
             False, ["0000316"], ['0000316', '0000167']),
            ('nm_url', "http://purl.obolibrary.org/obo/SO_0000167", False, [],
             ['0000167']),
            ('nm_url', "http://purl.obolibrary.org/obo/SO_0000139", False,
             ["0000316", '0000167'], ['0000139', '0000316', '0000167']),
            ('nm_url', "http://purl.obolibrary.org/obo/SO_000013", True, ["0000316"], ValueError),
            ('nm_url', "random stuff", True, ["0000316"], ValueError),
            ('nm_url', 576, True, [], TypeError)
        ]
    )
    def test_sbol_role(self, nm_url, cell_val, raising_err, roles,
                       expected, sbol_component, sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)
        if len(roles) > 0:
            for role in roles:
                sbol_component.roles = sbol_component.roles + [f'http://purl.obolibrary.org/obo/SO_{role}']

        if raising_err:
            with pytest.raises(expected):
                x.sbol_role()

        else:
            expected = [f'http://purl.obolibrary.org/obo/SO_{role}' for role in expected]
            x.sbol_role()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions['http://examples.org/test_comp_name/1']
            assert set(comp.roles) == set(expected)

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, expected',
        [
            ('nm_url', "4932", False, 'https://identifiers.org/taxonomy:4932'),
            ('nm_url', 4932, False, 'https://identifiers.org/taxonomy:4932'),
            ('nm_url', "23", False, 'https://identifiers.org/taxonomy:23'),
            ('nm_url', 23, False, 'https://identifiers.org/taxonomy:23'),
            ('nm_url', 'Hello', True, ValueError),
            ('nm_url', 1.5, False, 'https://identifiers.org/taxonomy:1'),
            ('nm_url', 1.0, False, 'https://identifiers.org/taxonomy:1'),
            ('nm_url', 1, False, 'https://identifiers.org/taxonomy:1'),
            ('nm_url', True, True, TypeError)
        ]
    )
    def test_sbh_targetOrganism(self, nm_url, cell_val, raising_err, expected,
                                sbol_component, sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)

        if raising_err:
            with pytest.raises(expected):
                x.sbh_targetOrganism()

        else:
            x.sbh_targetOrganism()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions['http://examples.org/test_comp_name/1']
            assert comp.targetOrganism == expected

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, expected',
        [
            ('nm_url', "4932", False, 'https://identifiers.org/taxonomy:4932'),
            ('nm_url', 4932, False, 'https://identifiers.org/taxonomy:4932'),
            ('nm_url', "23", False, 'https://identifiers.org/taxonomy:23'),
            ('nm_url', 23, False, 'https://identifiers.org/taxonomy:23'),
            ('nm_url', 'Hello', True, ValueError),
            ('nm_url', 1.5, False, 'https://identifiers.org/taxonomy:1'),
            ('nm_url', 1.0, False, 'https://identifiers.org/taxonomy:1'),
            ('nm_url', 1, False, 'https://identifiers.org/taxonomy:1'),
            ('nm_url', True, True, TypeError)
        ]
    )
    def test_sbh_sourceOrganism(self, nm_url, cell_val, raising_err, expected,
                                sbol_component, sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)

        if raising_err:
            with pytest.raises(expected):
                x.sbh_sourceOrganism()

        else:
            x.sbh_sourceOrganism()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions['http://examples.org/test_comp_name/1']
            assert comp.sourceOrganism == expected

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, pubmedExpected, expected',
        [
            ('nm_url', "https://pubmed.ncbi.nlm.nih.gov/24295448/", False,
             True, 'https://pubmed.ncbi.nlm.nih.gov/24295448'),
            ('nm_url', "http://parts.igem.org/Part:BBa_K2273000", False, False,
             'http://parts.igem.org/Part:BBa_K2273000'),
            ('nm_url', "https://www.addgene.org/87906/", False, False,
             'https://www.addgene.org/87906'),
            ('nm_url', "https://www.ncbi.nlm.nih.gov/nuccore/M11180.2", False,
             False, 'https://www.ncbi.nlm.nih.gov/nuccore/M11180.2'),
            ('nm_url', 'Hello', True, False, ValueError),
            ('nm_url', 1.5, True, False, TypeError),
            ('nm_url', 10, True, False, TypeError),
            ('nm_url', True, True, False, TypeError)
        ]
    )
    def test_sbh_dataSource(self, nm_url, cell_val, raising_err,
                            pubmedExpected, expected, sbol_component,
                            sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)

        if raising_err:
            with pytest.raises(expected):
                x.sbh_dataSource()

        else:
            x.sbh_dataSource()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions['http://examples.org/test_comp_name/1']
            if pubmedExpected:
                identi = os.path.split(expected)[1]
                assert comp.wasDerivedFrom[0] == expected and comp.OBI_0001617 == identi
            else:
                assert comp.wasDerivedFrom[0] == expected

    @pytest.mark.parametrize(
        'nm_url, cell_val, raising_err, expected',
        [
            ('nm_url', "https://synbiohub.org/public/Excel2SBOL/direct/1", False,
             'https://synbiohub.org/public/Excel2SBOL/direct/1'),
            ('nm_url', "https://synbiohub.org/public/Excel2SBOL/Unknown/1",
             False, 'https://synbiohub.org/public/Excel2SBOL/Unknown/1'),
            ('nm_url', "https://pubmed.ncbi.nlm.nih.gov/24295448/",
             True, ValueError),
            ('nm_url', 'Hello', True, ValueError),
            ('nm_url', 1.5, True, TypeError),
            ('nm_url', 10, True, TypeError),
            ('nm_url', True, True, TypeError)
        ]
    )
    def test_sbh_alteredSequence(self, nm_url, cell_val, raising_err,
                                 expected, sbol_component, sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)

        if raising_err:
            with pytest.raises(expected):
                x.sbh_alteredSequence()

        else:
            x.sbh_alteredSequence()
            sbol_doc.addComponentDefinition(sbol_component)

            comp = sbol_doc.componentDefinitions['http://examples.org/test_comp_name/1']
            assert comp.wasGeneratedBy[0] == expected

    @pytest.mark.parametrize(
        'nm_url, sbol_term, cell_val, raising_err, expected',
        [
            ('https://wiki.synbiohub.org/wiki/Terms/synbiohub#',
             'sbh_designNotes', 'No Kex recognition site', False,
             'No Kex recognition site'),
            ('https://wiki.synbiohub.org/wiki/Terms/synbiohub#',
             'sbh_designNotes', 1.5, False, "1.5"),
            ('https://wiki.synbiohub.org/wiki/Terms/synbiohub#',
             'sbh_designNotes', "12 test_näme%", False, "12 test_näme%"),
            ('http://sbols.org/v2#', 'sbol_roleCircular', True, False, "True")
        ]
    )
    def test_add_new(self, nm_url, sbol_term, cell_val, raising_err,
                     expected, sbol_component, sbol_doc):

        x = cf.sbol_methods(nm_url, sbol_component, sbol_doc, cell_val)
        x.sbol_term = sbol_term

        if raising_err:
            with pytest.raises(expected):
                x.add_new()

        else:
            x.add_new()
            sbol_doc.addComponentDefinition(sbol_component)
            sbol_term_sfx = sbol_term.split("_", 1)[1]

            comp = sbol_doc.componentDefinitions['http://examples.org/test_comp_name/1']
            comp_attr = getattr(comp, sbol_term_sfx)
            assert comp_attr == expected


@pytest.mark.parametrize(
    'column_dict_entry, raising_err, expected',
    [
        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'FALSE', 'Replacement Lookup': 'FALSE',
             'Sheet Name': '', 'From Col': '', 'To Col': ''}, False, 'no_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'else', 'Replacement Lookup': 'FALSE',
             'Sheet Name': '', 'From Col': '', 'To Col': ''}, True, TypeError
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': False, 'Replacement Lookup': 'FALSE',
             'Sheet Name': '', 'From Col': '', 'To Col': ''}, False, 'no_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'FALSE', 'Replacement Lookup': False,
             'Sheet Name': '', 'From Col': '', 'To Col': ''}, False, 'no_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': False, 'Replacement Lookup': False,
             'Sheet Name': '', 'From Col': '', 'To Col': ''}, False, 'no_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': 'TRUE',
             'Sheet Name': 'Replacement', 'From Col': 'A', 'To Col': 'B'},
            False, 'replacement_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': True, 'Replacement Lookup': 'TRUE',
             'Sheet Name': 'Replacement', 'From Col': 'A', 'To Col': '2'},
            False, 'replacement_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': True,
             'Sheet Name': 'Replacement', 'From Col': 'a', 'To Col': 2}, False,
            'replacement_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': True, 'Replacement Lookup': True,
             'Sheet Name': 'Replacement', 'From Col': 'A', 'To Col': 'B'},
            False, 'replacement_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': 'FALSE',
             'Sheet Name': 'Simple', 'From Col': 'A', 'To Col': 'B'}, False,
            'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': True, 'Replacement Lookup': 'FALSE',
             'Sheet Name': 'Simple', 'From Col': 'A', 'To Col': 'B'}, False,
            'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': False,
             'Sheet Name': 'Simple', 'From Col': 'A', 'To Col': 'B'}, False,
            'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': True, 'Replacement Lookup': False,
             'Sheet Name': 'Simple', 'From Col': 'A', 'To Col': 'B'}, False,
            'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': 'FALSE',
             'Sheet Name': 'NA_Between', 'From Col': 'A', 'To Col': 'D'},
            False, 'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': 'FALSE',
             'Sheet Name': 'NA_Reverse', 'From Col': 'D', 'To Col': 'A'},
            False, 'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': 'FALSE',
             'Sheet Name': 'Simple_Reverse', 'From Col': 'B', 'To Col': 'A'},
            False, 'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': 'FALSE',
             'Sheet Name': 'Row_3_Start', 'From Col': 'C', 'To Col': 'B'},
            False, 'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': 'FALSE',
             'Sheet Name': 'Blank col between', 'From Col': 'C',
             'To Col': 'A'}, False, 'norm_dict'
        ),

        (
            {'SBOL Term': 'sbol_term', 'Namespace URL': 'nm_url',
             'Sheet Lookup': 'TRUE', 'Replacement Lookup': 'FALSE',
             'Sheet Name': 'Doesnt exist', 'From Col': 'A', 'To Col': 'B'},
            True, ValueError
        )
    ]
)
def test_column_class(column_dict_entry, raising_err, expected,
                      norm_dict, replacement_dict):
    file_dir = os.path.dirname(__file__)
    file_path_in = os.path.join(file_dir, 'test_files',
                                'Column_Class_test.xlsx')
    if raising_err:
        with pytest.raises(expected):
            cf.column(file_path_in, column_dict_entry)

    else:
        x = cf.column(file_path_in, column_dict_entry)

        if expected == 'no_dict':
            assert x.sbol_term == 'sbol_term'
            assert x.namespace_url == 'nm_url'
            assert x.replacement_lookup is False
            with pytest.raises(AttributeError):
                x.lookup_dict
        elif expected == 'norm_dict':
            assert x.sbol_term == 'sbol_term'
            assert x.namespace_url == 'nm_url'
            assert x.replacement_lookup is False
            assert x.lookup_dict == norm_dict

        elif expected == 'replacement_dict':
            assert x.sbol_term == 'sbol_term'
            assert x.namespace_url == 'nm_url'
            assert x.replacement_lookup is True
            assert x.lookup_dict == replacement_dict



import tyto
import re
import excel_sbol_utils.helpers as hf


def up(col_def_row, cell_val, compiled_sheets, obj_dict):
    col_def_dict = col_def_row.to_dict(orient='list')
    if col_def_dict['Tyto Lookup'][0] and not col_def_dict['Sheet Lookup'][0]:
        # if the ontology lookup is TRUE and sheet lookup is FALSE
        # For returning the URI, we need the following:
        # ontology_name & cell_val
        er_val = cell_val
        onto_name = col_def_dict['Ontology Name'][0]
        if onto_name == "SO":
            cell_val = re.sub("[^A-Za-z0-9]", "_", cell_val)
        cell_val = tyto.endpoint.Ontobee.get_uri_by_term(getattr(tyto, onto_name), cell_val)
        if cell_val is None:
            raise ValueError(f'The Cell value {er_val} does not appear to be in {onto_name} please check spelling, capitalisation, and for species if it is the most up to date species name')
    if col_def_dict['Sheet Lookup'][0] and not col_def_dict['Replacement Lookup'][0]:
        # pull converted cell value from lookup table
        # created by table class and column class
        # and use the lookup column to get the new cell_value

        lk_dict_name = col_def_dict['Lookup Sheet Name'][0]
        lk_dict = compiled_sheets[lk_dict_name]['library']

        lk_col_from = list(lk_dict.keys())[hf.col_to_num(col_def_dict['From Col'][0]) - 1]
        lk_col_to = list(lk_dict.keys())[hf.col_to_num(col_def_dict['To Col'][0]) - 1]

        try:
            # find the index of cell value in the from column
            # find the item with the same index in the to column
            cell_val = lk_dict[lk_col_to][lk_dict[lk_col_from].index(cell_val)]
        except ValueError:
            raise KeyError(f'cell vlaue: {cell_val} not in the lookup dictionary: {lk_dict_name}')
    elif col_def_dict['Sheet Lookup'][0]:
        # if it is a lookup and a replacement lookup
        # create a url based on the prefix
        # E.g. pubmed:1023 means use pubmed url and value 1023
        cell_val_prefix = cell_val.split(":", 1)[0]
        cell_val_suffix = cell_val.split(":", 1)[1]

        lk_dict_name = col_def_dict['Lookup Sheet Name'][0]
        lk_dict = compiled_sheets[lk_dict_name]['library']

        lk_col_from = list(lk_dict.keys())[hf.col_to_num(col_def_dict['From Col'][0]) - 1]
        lk_col_to = list(lk_dict.keys())[hf.col_to_num(col_def_dict['To Col'][0]) - 1]

        try:
            # find the index of cell value in the from column
            # find the item with the same index in the to column
            cell_val = lk_dict[lk_col_to][lk_dict[lk_col_from].index(cell_val_prefix)].replace("{REPLACE_HERE}", cell_val_suffix)
        except ValueError:
            raise KeyError(f'cell vlaue: {cell_val_prefix} not in the lookup dictionary: {lk_dict_name}')

    elif col_def_dict['Object_ID Lookup'][0] and not col_def_dict['Parent Lookup'][0]:
        # changes part name to the uri to reference (if not a parental lookup)
        # could later add an if statement to only go through conversion
        # if the cell value is not already a uri, this would allow subcomponents
        # to be referenced either by part id if built in sheet or by uri
        # if the component already exists
        try:
            cell_val = obj_dict[cell_val]['uri']
        except KeyError:
            raise KeyError(f'The object "{cell_val}" is referenced in Sheet: "{col_def_dict["Sheet Name"][0]}", Column: "{col_def_dict["Column Name"][0]}" but is never created')
    return cell_val



import excel2sbol.converter as conv
import os

cwd = os.getcwd()
print(cwd)

file_path_in = os.path.join(cwd, "excel2sbol", 'tests', 'test_files',
                            'SBOL2_parts_project.xlsx')
file_path_out = os.path.join(cwd, "excel2sbol", 'tests', 'test_files',
                             'SBOL2_parts_project.xml')

conv.converter(file_path_in, file_path_out)
# conv.converter(file_path_in, file_path_out, sbol_version=2,
#                homespace="http://examples.org/")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 25 16:41:00 2020

@author: isapoetzsch
"""

from setuptools import find_packages, setup

setup(name='excel2sbol',
      version='1.0.32',
      url='https://github.com/SynBioDex/Excel-to-SBOL/',
      license='BSD 3-clause',
      maintainer='Jet Mante',
      maintainer_email='jet.mante@colorado.edu',
      include_package_data=True,
      description='convert excel resources into sbol',
      packages=find_packages(include=['excel2sbol']),
      long_description=open('README.md').read(),
      install_requires=['sbol2>=1.4.1',
                        'pandas>=1.0.1',
                        'numpy>=1.18.1',
                        'validators>=0.18.2',
                        'openpyxl>=3.0.7',
                        'tyto>=1.2',
                        'sbol3>=1.0.1',
                        'xlrd >= 1.0.0',
                        'excel_sbol_utils >= 1.0.42'],
      zip_safe=False)



# run by typing 'pytest' into the terminal
# for more detailed output use 'pytest -v -s'
import pytest
import excel2sbol.helper_functions as hf


class Test_col_to_num:

    @pytest.mark.parametrize(
        'col_name, raising_err, expected', [
            ("A", False, 0),
            ("a", False, 0),
            ("aa", False, 26),
            ("AA", False, 26),
            ("B", False, 1),
            ("Not a column", True, ValueError),
            ("LONG", True, ValueError),
            (3, True, TypeError)
        ]
    )
    def test_col_to_num(self, col_name, raising_err, expected):
        if raising_err:
            with pytest.raises(expected):
                hf.col_to_num(col_name)
        else:
            assert hf.col_to_num(col_name) == expected


class Test_check_name:

    @pytest.mark.parametrize(
        'nm_to_chck, raising_err, expected', [
            ("testname", False, "testname"),
            ("test_name", False, "test_name"),
            ("_test_name", False, "_test_name"),
            ("test_name12", False, "test_name12"),
            ("12test_name", False, "_12test_name"),
            ("test_name%", False, "test_name_u37_"),
            ("test_näme", False, "test_n_u228_me"),
            ("tεst_name", False, "t_u949_st_name"),
            (576, True, TypeError)
        ]
    )
    def test_check_name(self, nm_to_chck, raising_err, expected):
        if raising_err:
            with pytest.raises(expected):
                hf.check_name(nm_to_chck)
        else:
            assert hf.check_name(nm_to_chck) == expected


class Test_truthy_strings:
    @pytest.mark.parametrize(
        'to_chck, raising_err, expected', [
            ("True", False, True),
            ("TRUE", False, True),
            ("true", False, True),
            ("tRue", False, True),
            (True, False, True),
            ("False", False, False),
            ("FALSE", False, False),
            ("false", False, False),
            ("fAlse", False, False),
            (False, False, False),
            (576, True, TypeError),
            ("something", True, TypeError),
            ("1RUE", True, TypeError)
        ]
    )
    def test_check_name(self, to_chck, raising_err, expected):
        if raising_err:
            with pytest.raises(expected):
                hf.truthy_strings(to_chck)
        else:
            assert hf.truthy_strings(to_chck) == expected
